package javax.rtgedition;

/* AbsoluteTime.java
 * Represents AbsoluteTime in canonical Form
 * 
 * @author Vito J. Sabella
 * @version 1.0.0
 */

import java.util.Date;

public class AbsoluteTime extends HighResolutionTime
{
    /* Equivalant to AbsoluteTime(0,0) */
    public AbsoluteTime()
    {
        this(0,0);
    }

  /* NOTE! If you give me a null time I'll give you a null pointer exception. 
     So don't mess with me! :-) -vjs Nov 09 2001
  */
  
  public AbsoluteTime(AbsoluteTime time)
  {
      this.millis = time.millis;
      this.nanos  = time.nanos;
  }

  public AbsoluteTime(java.util.Date date)
  {
      millis = date.getTime();
      nanos = 0;   
  }

  public AbsoluteTime(long millis, int nanos)
  {
      this.millis = millis+nanos/NANOS_TO_MILLIS;
      this.nanos = nanos%NANOS_TO_MILLIS;
  }

  /* Public Methods */
  
  
  
  public AbsoluteTime absolute(Clock clock, AbsoluteTime destination)
  {
      /*
       * The specification states that we should pretend that this absoluteTime
       * is already referenced to the given clock so no check is needed.
      */
 
      if(destination == null)
      {
          return this;
      }
      destination.set(this);
      return destination;
  }

  public AbsoluteTime add(long millis, int nanos)
  {   
      /*
       * Note constructors are compensated for canonical time.
       */
      
      return new AbsoluteTime(this.millis+millis, this.nanos+nanos);
  }

  public AbsoluteTime add(long millis, int nanos, AbsoluteTime destination)
  {
      if(destination == null)
      {
          return new AbsoluteTime(this.millis+millis,this.nanos+nanos);
          /*
           * Note this is to save the clock cycles of a destination.set();
           * Also this is proper behavior.  Java is pass by value
           * EVEN for heap-allocated objects.  (Pointers are copies.)
           */
      }
      
      
    /*   Note, destination.set(long,int) calls 
         HighResolutionTime.set(long,int) which is compensated for 
         canonical time!
    */
      
      destination.set(this.millis+millis,this.nanos+nanos);
      return destination;
  }


  /*
   *  If the RelativeTime is null we throw a null pointer exception.
   *  We should only make a new object if destination is null but that
   *  makes sense.  No point in compensating for programmer error and making
   *  it failsafe if RelativeTime time is null.  Follow rule for entire class.
   *  - Modification, VJS Nov 10 2001
   */
  
  public final AbsoluteTime add(RelativeTime time)
  {   
      return new AbsoluteTime(this.millis+time.millis,this.nanos+time.nanos);
  }

  
  public AbsoluteTime add(RelativeTime time, AbsoluteTime destination)
  {
	if(destination == null)
        {
            return new AbsoluteTime(this.millis+time.millis,this.nanos+time.nanos);
	}
        
        /* Note, destination.set(long,int) calls 
           HighResolutionTime.set(long,int) which is compensated for 
           canonical time!
         */
        
	destination.set(this.millis+time.millis,this.nanos+time.nanos);
        
        return destination;
  }

  public java.util.Date getDate()
  {
      return new Date(this.millis);
  }

  public void set(java.util.Date date)
  {
      millis = date.getTime();
      nanos = 0;
  }

  /*
   * Note - a real spec writer :-) would have included an InvalidTimeException
   * if this were null.  But exception tracking takes up more system resources.
   */
  
  public final RelativeTime subtract(AbsoluteTime time)
  {
      int nanos = this.nanos-time.nanos;
      long millis = this.millis-time.millis;
      
      /*
       * Since nanos is in canonical form, subtraction will
       * always be less than 1e6 (1ms).
       * BUT if nanos ends up being negative we need to borrow
       * a millisecond;
       */
      
      
      if(nanos < 0)
      {
          millis -= 1;
          nanos  = nanos+NANOS_TO_MILLIS;
      }
      
      return new RelativeTime(millis,nanos);
  }
 
  
  /*
   * Puts result in destination and returns result.  If destination is null
   * makes a new RelativeTime object for it.
   *
   * @see subtract(AbsoluteTime time);
   */
  
  public RelativeTime subtract(AbsoluteTime time, RelativeTime destination)
  {
      if(destination == null)
      {
          destination = new RelativeTime();
      }
      
      int nanos = this.nanos-time.nanos;
      long millis = this.millis-time.millis;
      
      if(nanos < 0)
      {
          millis -= 1;
          nanos  = nanos+NANOS_TO_MILLIS;
      }
      
      destination.set(millis,nanos);
      return destination;
  }
  
  public final AbsoluteTime subtract(RelativeTime time)
  {   
      //Rember: As above, time is already canonical.
      
      int  nanos  = this.nanos-time.nanos;
      long millis = this.millis-time.millis;
      
      if(nanos < 0)
      {
          millis -= 1;
          nanos   = nanos+NANOS_TO_MILLIS;
      }
      
      return new AbsoluteTime(millis,nanos);
  }
  
  /*
   * Puts result in destination and returns result.  If destination is null
   * makes a new AbsoluteTime object for it.
   *
   * @see subtract(RelativeTime time);
   */
  
  public AbsoluteTime subtract(RelativeTime time, AbsoluteTime destination)
  {
      if(destination == null)
      {
          destination = new AbsoluteTime();
      }
      
      int  nanos  = this.nanos-time.nanos;
      long millis = this.millis-time.millis;
      
      if(nanos < 0)
      {
          millis -= 1;
          nanos   = nanos+NANOS_TO_MILLIS;
      }
      
      destination.set(millis,nanos);
      return destination;
  }
  
  public String toString()
  {
      return (this.getDate()).toString() + " : "+ this.millis + "ms "+this.nanos+"ns";   
  }

  
} 
