/*
 * Copyright (c) 1998-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Event handling support
 * FILE:      events.h
 * OVERVIEW:  This file defines the macros and operations for
 *            binding the interpreter to the event handling
 *            mechanisms of the host operating system in a 
 *            portable fashion.
 * AUTHOR:    Nik Shaylor 4/20/00
 *            Original Palm OS specific implementation by Doug Simon 1998
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

/*=========================================================================
 * Event system macros
 *=======================================================================*/

/*=========================================================================
 * Thread scheduling macros.
 * These are here to isolate the scheduling code out of interpret.c.
 * Note that the thread switching macros are normally used also for
 * driving the event handling/polling mechanism,
 *=======================================================================*/

/*
 * Indicate task switching is necessary
 * (enforce a thread switch)
 */
#define signalTimeToReschedule() (Timeslice = 0)

/*
 * Determine if it is time to reschedule
 * (by decrementing the time slice and comparing
 * it to zero)
 */
#define isTimeToReschedule() (Timeslice-- == 0)

/*
 * Check that the time slice value is not junk
 * (should never happen)
 */
#if INCLUDEDEBUGCODE
#define checkRescheduleValid() if (Timeslice < 0) \
    { fatalError(KVM_MSG_INVALID_TIMESLICE); }
#else
#define checkRescheduleValid() /**/
#endif /* INCLUDEDEBUGCODE */

/*
 * BetterHandleEvent
 *
 * This function ties the virtual machine with events coming
 * from the host operating system.  Since CLDC itself does 
 * not define any event handling capabilities, the function
 * pretty much empty in those builds that don't include any
 * additional classes beyond CLDC.
 *
 */
#if INCLUDE_ALL_CLASSES
void BetterHandleEvent(ulong64);
#else
#define BetterHandleEvent(t)
#endif /* INCLUDE_ALL_CLASSES */

/*
 * __ProcessDebugCmds()
 *
 * This function ties the virtual machine with the new
 * Java-level debugger interface.
 */
#if ENABLE_JAVA_DEBUGGER
#define __ProcessDebugCmds(x) ProcessDebugCmds(x);
#else
#define __ProcessDebugCmds(x)
#endif /* ENABLE_JAVA_DEBUGGER */

/*
 * Reschedule
 *
 * This routine is called from inside the interpreter when
 * it is time to perform thread switching.
 *
 */
#define reschedule()                                                    \
        do  {                                                           \
            ulong64 wakeupTime;                                         \
            if (!areAliveThreads()) {                                   \
                return;   /* end of program */                          \
            }                                                           \
            checkTimerQueue(&wakeupTime);                               \
            BetterHandleEvent(wakeupTime);                              \
            __ProcessDebugCmds(0);                                      \
        } while (!SwitchThread());


/*=========================================================================
 * Data structures needed for event handling
 *=======================================================================*/

/*
 * These macros are used as parameters to the power conservation
 * features of the KVM.
 */
#define FOREVER TRUE
#define FIXED_DELAY FALSE

/*
 *  The type of events for which handlers can be registered. Corresponds
 *  exactly with the handler methods defined in com.sun.kjava.Spotlet.
 *  The "pen*" constants must be the first three elements in the list
 *  and in the order DOWN,UP,MOVE to correspond with the definition
 *  in the events enumerated type in Events.h.
 */
enum KVMEventTypes {
    beamReceiveKVMEvent = 0,
    penDownKVMEvent = 1,
    penUpKVMEvent   = 2,
    penMoveKVMEvent = 3,
    keyDownKVMEvent = 4,
    lastKVMEvent =    5,
    appStopKVMEvent = 5
};

bool_t HandleEvent(ulong64 nextTimerEvent);

/*  The event record. */
typedef struct {
    enum KVMEventTypes type;
    BYTEARRAY      beam;
    short          screenX;
    short          screenY;
    short          chr;     /* Palm has 16-bit chars */
} KVMEventType;

/* The option flags when registering for events. */
enum eventOptions {
    NOEVENTOPTIONS = 0x00,
    WANTSYSTEMKEYS = 0x01   /* User wants system "hard" key events */
};

/*=========================================================================
 * Event handling functions
 *=======================================================================*/

void InitializeEvents(void);
void storeEvent(KVMEventType*);

/*
 * GetNextKVMEvent
 *
 * Platform-specific function for reading events from the host
 * operating system.  This function typically needs to be ported 
 * to use the low-level event handling mechanisms of the host
 * platform.
 */
bool_t
GetNextKVMEvent(KVMEventType *event, bool_t forever, ulong64 waitUntil);


