/*
 *  Copyright (c) 1998-2001 Sun Microsystems, Inc., 901 San Antonio Road,
 *  Palo Alto, CA 94303, U.S.A.  All Rights Reserved.
 *
 *  Sun Microsystems, Inc. has intellectual property rights relating
 *  to the technology embodied in this software.  In particular, and
 *  without limitation, these intellectual property rights may include
 *  one or more U.S. patents, foreign patents, or pending
 *  applications.  Sun, Sun Microsystems, the Sun logo, Java, KJava,
 *  and all Sun-based and Java-based marks are trademarks or
 *  registered trademarks of Sun Microsystems, Inc.  in the United
 *  States and other countries.
 *
 *  This software is distributed under licenses restricting its use,
 *  copying, distribution, and decompilation.  No part of this
 *  software may be reproduced in any form by any means without prior
 *  written authorization of Sun and its licensors, if any.
 *
 *  FEDERAL ACQUISITIONS:  Commercial Software -- Government Users
 *  Subject to Standard License Terms and Conditions
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Event handling support
 * FILE:      events.c
 * OVERVIEW:  This file defines the operations for binding the
 *            interpreter to the event handling mechanisms of
 *            the host operating system.
 * AUTHOR:    Nik Shaylor 4/20/00
 *            Bill Pittore (debugging support)
 *=======================================================================*/

/*
 * This file implements the "events:" protocol
 *
 * @author  Nik Shaylor
 * @version 1.0 1/15/2000
 */

#define MAXPARMLENGTH 20

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>
#if ENABLE_JAVA_DEBUGGER
#include <debugger.h>
#endif

/*=========================================================================
 * Local variables
 *=======================================================================*/

static THREAD waitingThread;
static int    opened;
static cell   parameter[MAXPARMLENGTH];
static int    count;
static int    eventInP;
static int    eventOutP;

/*=========================================================================
 * Event handling functions
 *=======================================================================*/

/*=========================================================================
 * FUNCTION:      InitializeEvents
 * TYPE:          global function
 * OVERVIEW:      Initialize the event system.
 *                We have to explicitly initialize the event system
 *                so that the KVM can shutdown and restart cleanly.
 * INTERFACE:
 *   parameters:  none
 *   returns:     none
 *=======================================================================*/

void InitializeEvents()
{
    waitingThread = 0;
    makeGlobalRoot((cell **)&waitingThread);
    opened    = FALSE;
    count     = 0;
    eventInP  = 0;
    eventOutP = 0;

#if INCLUDEDEBUGCODE
    if (traceevents) {
        fprintf(stdout, "Event system initialized\n");
    }
#endif /* INCLUDEDEBUGCODE */

}

static cell popParameter()
{
    cell tmpParam = 0;
    
    if (eventOutP < MAXPARMLENGTH) {
        tmpParam = parameter[eventOutP++];
    }
    /* if we consumed all events, reset pointers */
    if (eventOutP == eventInP)
        eventInP = eventOutP = 0;
    return (tmpParam);
}

/*
 * The maximum we push is 3 values so check for enough room for a whole
 * event
 */
#define PUSHPARAMETER(param)                            \
{                                                       \
        if (eventInP < (MAXPARMLENGTH - 3)) {           \
            parameter[eventInP++] = param;              \
            count ++;                                   \
        }                                               \
}

void storeEvent(KVMEventType *event) {

        PUSHPARAMETER((cell)event->type)

        switch (event->type) {
            case penDownKVMEvent:
            case penUpKVMEvent:
            case penMoveKVMEvent:
                PUSHPARAMETER((cell)event->screenX)
                PUSHPARAMETER((cell)event->screenY)
                return;

            case keyDownKVMEvent:
                PUSHPARAMETER((cell)event->chr)
                return;

            case beamReceiveKVMEvent:
                PUSHPARAMETER((cell)event->beam)
                return;

            default:; /* do nothing */
        }
}

/*=========================================================================
 * FUNCTION:      getEvent
 * TYPE:          local function
 * OVERVIEW:      Wait for a system event
 * INTERFACE:
 *   parameters:  time to wait if no events are forthcoming;
 *                parameter value 0 indicates that the host OS
 *                can wait (and conserve battery) until
 *                new native events are available.
 *   returns:     flag to say if an event was found
 *
 * NOTE:          This implementation allows a KVM port to
 *                take advantage of machine-specific battery
 *                conservation features.  If the "wakeupTime"
 *                parameter is zero, the "forever" parameter is
 *                set to TRUE.  This is an indication to the platform-
 *                specific implementation of "GetNextKVMEvent()" that
 *                the host platform can sleep/conserve battery until
 *                the next native event occurs.
 *=======================================================================*/

static int getEvent(ulong64 wakeupTime) {
    KVMEventType event;
    bool_t forever;

    /* We wait forever if there is nothing else to run */
    if (!areActiveThreads()) {

#if ENABLE_JAVA_DEBUGGER
        forever = FALSE;
        ll_setZero(wakeupTime);
#else
        /* Set 'forever' to true if wakeupTime is zero */
        forever = ll_zero_eq(wakeupTime);
#endif

    } else {
        forever = FALSE;
        ll_setZero(wakeupTime);
    }

    while (TRUE) {

        /* Get the next available event */
        /* Note: the platform-specific implementation */
        /* of GetNextKVMEvent should instruct the host platform */
        /* to utilize platform-specific battery conservation */
        /* functions when "forever == TRUE and wakeupTime == 0" */
        if (!GetNextKVMEvent(&event, forever, wakeupTime)) {        
            return FALSE;
        }

#if INCLUDEDEBUGCODE
        if (traceevents) {
            sprintf(str_buffer, "Event %ld received\n", (long)event.type);
            fprintf(stdout, str_buffer);
        }
#endif /* INCLUDEDEBUGCODE */

        storeEvent(&event);

        switch (event.type) {
            case penDownKVMEvent:
            case penUpKVMEvent:
            case penMoveKVMEvent:
            case keyDownKVMEvent:
            case beamReceiveKVMEvent:
                return TRUE;

            case appStopKVMEvent:
                ERROR_THROW(0);
                return TRUE;

            default:; /* do nothing, but continue in loop */
        }
    }
}

 /*=========================================================================
  * FUNCTION:      BetterHandleEvent
  * TYPE:          Global C function
  * OVERVIEW:      Wait for a system event
  * INTERFACE:
  *   parameters:  time to wait if no events are available immediately;
  *                zero value indicates that the host platform can go
  *                to sleep and/or conserve battery until a new native
  *                event occurs
  *   returns:     nothing
  *=======================================================================*/

void BetterHandleEvent(ulong64 wakeupTime) {

    /* Check if event stream is open */
    if (opened) {

        /* Get next event, sleeping if necessary */
        if (waitingThread != 0 && getEvent(wakeupTime) ) {

            /* Store the received data in the operand stack */
            /* of the thread that is using the event stream */
            count--;
            (void)popStackForThread(waitingThread);
            pushStackForThread(waitingThread, popParameter());
            resumeThread(waitingThread);
            waitingThread = 0;
        }

    } else {

        /* Uncommon case: Event stream is not open. */
        /* Wait as long as necessary, but allow debug */
        /* operations to be processed */
        if (!areActiveThreads() && ll_zero_ne(wakeupTime)) {

#if ENABLE_JAVA_DEBUGGER
            for(;;) { 
                long64 delta64 = *(long64*)&wakeupTime;
                ll_dec(delta64, CurrentTime_md());
                if (ll_zero_gt(delta64)) { 
                    unsigned long delta;
                    ll_long_to_uint(delta64, delta);
                    ProcessDebugCmds(delta);
                } else { 
                    break;
                }
            }
#else
            /* Note: This will currently waste battery on the Palm.  */
            /* We don't care, since it makes no sense to run Java    */
            /* programs on the Palm without opening an event stream. */
            SLEEP_UNTIL(wakeupTime);

#endif /* ENABLE_JAVA_DEBUGGER */

        }
    }
}

/*=========================================================================
 * Protocol implementation methods ("events:" protocol)
 *=======================================================================*/

 /*=========================================================================
  * FUNCTION:      open()Z (VIRTUAL)
  * CLASS:         com.sun.cldc.io.j2me.events.Protocol
  * TYPE:          virtual native function
  * OVERVIEW:      Read an integer from the event queue
  * INTERFACE (operand stack manipulation):
  *   parameters:  this
  *   returns:     the integer value
  *=======================================================================*/

void Java_com_sun_cldc_io_j2me_events_PrivateInputStream_open(void) {
    INSTANCE instance = popStackAsType(INSTANCE);
    (void)instance;
    if (opened) {
        raiseException("java/lang/IllegalAccessException");
    }
    opened = TRUE;
}

 /*=========================================================================
  * FUNCTION:      close()Z (VIRTUAL)
  * CLASS:         com.sun.cldc.io.j2me.events.Protocol
  * TYPE:          virtual native function
  * OVERVIEW:      Read an integer from the event queue
  * INTERFACE (operand stack manipulation):
  *   parameters:  this
  *   returns:     the integer value
  *=======================================================================*/

void Java_com_sun_cldc_io_j2me_events_PrivateInputStream_close(void) {
    INSTANCE instance = popStackAsType(INSTANCE);
    (void)instance;
    if (!opened) {
        raiseException("java/lang/IllegalAccessException");
    }
    opened = FALSE;
}

 /*=========================================================================
  * FUNCTION:      readInt0()Z (VIRTUAL)
  * CLASS:         com.sun.cldc.io.j2me.events.Protocol
  * TYPE:          virtual native function
  * OVERVIEW:      Read an integer from the event queue
  * INTERFACE (operand stack manipulation):
  *   parameters:  this
  *   returns:     the integer value
  *=======================================================================*/

void Java_com_sun_cldc_io_j2me_events_PrivateInputStream_readInt(void) {
    INSTANCE instance = popStackAsType(INSTANCE);
    (void)instance;
    if (count == 0) {
        waitingThread = CurrentThread;
        pushStack(0);  /* For Dear Uncle Frank's checker */
        suspendThread();
    } else {
        --count;
        pushStack(popParameter());
    }
 }

 /*=========================================================================
  * FUNCTION:      readByteArray()Z (VIRTUAL)
  * CLASS:         com.sun.cldc.io.j2me.events.Protocol
  * TYPE:          virtual native function
  * OVERVIEW:      Read byte array from the event queue
  * INTERFACE (operand stack manipulation):
  *   parameters:  this
  *   returns:     the byte array
  *=======================================================================*/

void Java_com_sun_cldc_io_j2me_events_PrivateInputStream_readByteArray(void) {
    INSTANCE instance = popStackAsType(INSTANCE);
    (void)instance;
    if (count == 0) {
        raiseException("java/lang/IllegalAccessException");
    } else {
        --count;
        pushStack(popParameter());
    }
}

