/*
 * Copyright (c) 1998-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Profiling
 * FILE:      log.c
 * OVERVIEW:  Operations for gathering and printing out diagnostic
 *            virtual machine execution information at runtime.
 * AUTHOR:    Daniel Blaukopf, based on code by Antero Taivalsaari
 *            and Doug Simon
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>

/*=========================================================================
 * Functions
 *=======================================================================*/

/*=========================================================================
 * NOTE:
 * Documentation for each of the functions below is provided in log.h.
 *=======================================================================*/

static void log_nullFunction() { return; }

#if INCLUDEDEBUGCODE

static void log_enterMethod(METHOD method, const char *type) 
{ 
    if (tracemethodcallsverbose) {
        ASSERTING_NO_ALLOCATION
            char className[256];
            char *methodName = methodName(method);
            getClassName_inBuffer((CLASS)method->ofClass, className);
            if (TERSE_MESSAGES) {
                fprintf(stdout, "=> %s.%s\n", className, methodName);
            } else {
                char methodSignature[256];
                unsigned short key = method->nameTypeKey.nt.typeKey;
                change_Key_to_MethodSignature_inBuffer(key, methodSignature);
                fprintf(stdout, "Invoking %s method '%s::%s%s'\n", 
                        type, className, methodName, methodSignature);
            }
        END_ASSERTING_NO_ALLOCATION
    }
}

#endif

/*=========================================================================
 * FUNCTION:      TraceMethodExit()
 * OVERVIEW:      Trace a method exit
 * INTERFACE:
 *   parameters:  METHOD method
 *   returns:     <nothing>
 *=======================================================================*/

#if INCLUDEDEBUGCODE
static void log_exitMethod(METHOD method) 
{
    if (tracemethodcallsverbose) {
        ASSERTING_NO_ALLOCATION
            char className[256];
            char *methodName = methodName(method);
            getClassName_inBuffer((CLASS)method->ofClass, className);
            if (TERSE_MESSAGES) {
                fprintf(stdout, "<= %s.%s\n", className, methodName);
            } else { 
                fprintf(stdout, "Leaving %s::%s\n", className, methodName);
            }
        END_ASSERTING_NO_ALLOCATION
    }
}
#else
#define log_enterMethod ((void (*)(METHOD, const char*))log_nullFunction)
#define log_exitMethod  ((void (*)(METHOD)             )log_nullFunction)
#endif

#if INCLUDEDEBUGCODE
static void log_allocateObject(long pointer,
                               long cells,
                               int type,
                               long ID,
                               long memoryFree) {
    if (tracememoryallocation) {
        fprintf(stdout, "Allocated %ld bytes, address: %lx, type: %lx, free: %ld\n",
            cells*CELL, pointer, (long)type, memoryFree);
    }
}

static void log_allocateHeap(long heapSize,
                             long heapBottom,
                             long heapTop) {
    if (tracememoryallocation) {
        fprintf(stdout, "Allocated a dynamic heap of %ld bytes\n", heapSize);
        fprintf(stdout, "Heap bottom: %lx, heap top: %lx\n",
            heapBottom, heapTop);
    }
}

#else /* INCLUDEDEBUGCODE */

#define log_allocateObject ((void (*)(long,long,int,long,long))log_nullFunction)
#define log_allocateHeap   ((void (*)(long,long,long))log_nullFunction)

#endif /* INCLUDEDEBUGCODE */

#if INCLUDEDEBUGCODE

static void log_heapObject(long object,
                           const char *type,
                           long size,
                           int isMarked,
                           const char *className,
                           int isClass) {
    if (tracegarbagecollectionverbose) {
        fprintf(stdout, "%lx: %s, size=%ld%s",
            object, type, size,
            isMarked ? ", MARKED" : "");
        if (className==NULL) {
            fprintf(stdout, "\n");
        } else if (isClass) {
            fprintf(stdout, ", class %s\n", className);
        } else {
            fprintf(stdout, ", instance of %s\n", className);
        }
    }
}

static void log_heapSummary(int objectCount, int freeCount,
                            int liveByteCount, int freeByteCount,
                            int largestFreeObjectSize,
                            long heapBottom, long heapTop) {
    if (tracegarbagecollectionverbose) {
        fprintf(stdout, "%d objects in heap (%d alive/%d free)\n",
            objectCount+freeCount, objectCount, freeCount);
        fprintf(stdout, "%d live bytes; %d bytes free (%d total)\n",
            liveByteCount, freeByteCount, liveByteCount+freeByteCount);
        fprintf(stdout, "Largest free object: %d bytes\n",
            largestFreeObjectSize);
        fprintf(stdout, "Heap bottom: %lx, heap top: %lx\n",
            heapBottom, heapTop);
    }
}

static void log_horizontalRule(void) {
    if (tracegarbagecollectionverbose) {
        fprintf(stdout, "\n====================\n");
    }
}

#define log_startHeapScan log_horizontalRule
#define log_endHeapScan log_horizontalRule

static void log_heapWarning (long freeListCount,
                             long byteCount) {
    if (tracegarbagecollectionverbose) {
        fprintf(stdout, "WARNING: %ld objects in the free list (%ld bytes)\n",
            freeListCount, byteCount);
    }
}


#else
#define log_heapObject \
  ((void (*)(long,const char*,long,int,const char*,int))log_nullFunction)
#define log_heapSummary \
  ((void (*)(int,int,int,int,int,long,long))log_nullFunction)
#define log_startHeapScan ((void (*)(void))log_nullFunction)
#define log_endHeapScan   ((void (*)(void))log_nullFunction)
#define log_heapWarning   ((void (*)(long,long))log_nullFunction)
#endif

#if INCLUDEDEBUGCODE
static void log_loadClass(const char *className) {
    if (traceclassloadingverbose) {
        fprintf(stdout, "Loading class '%s'\n", className);
    }
}
#else
#define log_loadClass ((void (*)(const char *))log_nullFunction)
#endif   
    
#if INCLUDEDEBUGCODE
static void 
log_throwException(THROWABLE_INSTANCE exception) {
    if (traceexceptions) {
        fprintf(stdout, "Exception: %s\n", \
        getClassName((CLASS)exception->ofClass));
    }
}
#else
#define log_throwException ((void(*)(THROWABLE_INSTANCE))log_nullFunction)
#endif /* INCLUDEDEBUGCODE */

static void 
log_uncaughtException(THROWABLE_INSTANCE exception) 
{
    fprintf(stderr, "Uncaught exception %s\n", \
    getClassName((CLASS)exception->ofClass));
}

static const struct KVMLogInterface_ logImplementation = {
    fprintf,
    log_enterMethod,
    log_exitMethod,
    log_allocateObject,
    log_allocateHeap,
    log_heapObject,
    log_heapSummary,
    log_startHeapScan,
    log_endHeapScan,
    log_heapWarning,
    log_loadClass,
    log_throwException,
    log_uncaughtException
};

const KVMLogInterface Log = (KVMLogInterface) &logImplementation;

