/*
 * Copyright (c) 1998-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Native function interface
 * FILE:      native.c
 * OVERVIEW:  This file defines the interface for plugging in
 *            the native functions needed by the Java Virtual
 *            Machine. The implementation here is _not_ based
 *            on JNI (Java Native Interface), because JNI seems
 *            too expensive for small devices.
 * AUTHOR:    Antero Taivalsaari, Sun Labs
 *            Edited by Doug Simon 11/1998
 *            Frank Yellin
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>

/*=========================================================================
 * Operations on native functions
 *=======================================================================*/

/*=========================================================================
 * FUNCTION:      getNativeFunction()
 * TYPE:          lookup operation
 * OVERVIEW:      Given a function name as a string, try to find
 *                a corresponding native function and return a
 *                a pointer to it if found.
 * INTERFACE:
 *   parameters:  class name, method name
 *   returns:     function pointer or NIL if not found.
 *=======================================================================*/

NativeFunctionPtr
getNativeFunction(INSTANCE_CLASS clazz, const char* methodName)
{
#if !ROMIZING
    const ClassNativeImplementationType *cptr;
    const NativeImplementationType *mptr;
    UString UBaseName    = clazz->clazz.baseName;
    UString UPackageName = clazz->clazz.packageName;
    char *baseName = UStringInfo(UBaseName);
    char *packageName = UStringInfo(UPackageName);
    for (cptr = nativeImplementations; ; cptr++) {
        if (cptr->packageName == NULL) { 
            return NULL;
        }
        if (   (strcmp(cptr->packageName, packageName) == 0) 
            && (strcmp(cptr->baseName,    baseName) == 0)) { 
            break;
        }
    }
    for (mptr = cptr->implementation; ; mptr++) {
        const char *name = mptr->name;
        if (name == NULL) {
            return NULL;
        }
        if (strcmp(name, methodName) == 0) {
            return mptr->implementation;
        }
    }

#else
    return NULL;
#endif /* !ROMIZING */
}

/*=========================================================================
 * FUNCTION:      invokeNativeFunction()
 * TYPE:          private operation
 * OVERVIEW:      Invoke a native function, resolving the native
 *                function reference if necessary.
 * INTERFACE:
 *   parameters:  method pointer
 *   returns:     <nothing>
 * NOTE:          Native functions are automatically synchronized, i.e.,
 *                they cannot be interrupted by other threads unless
 *                the native code happens to invoke the Java interpreter.
 *=======================================================================*/

void invokeNativeFunction(METHOD thisMethod)
{
    NativeFunctionPtr native = thisMethod->u.native.code;

    if (native == NULL) {
        /*  Native function not found; throw error */

        /* The GC may get confused by the arguments on the stack */
        setSP(getSP() - thisMethod->argCount);

        START_TEMPORARY_ROOTS
            DECLARE_TEMPORARY_ROOT(char*, className, 
                     getClassName((CLASS)(thisMethod->ofClass)));
            sprintf(str_buffer,
                    KVM_MSG_NATIVE_METHOD_NOT_FOUND_2STRPARAMS,
                    className, methodName(thisMethod));
        END_TEMPORARY_ROOTS
        fatalError(str_buffer);
    }

#if INCLUDEDEBUGCODE
        if (tracemethodcalls || tracemethodcallsverbose) {
            frameTracing(thisMethod, "=>", +1);
        }
#endif

    /* Call the native function we are supposed to call */
    native();

#if INCLUDEDEBUGCODE
        if (tracemethodcalls || tracemethodcallsverbose) {
            frameTracing(thisMethod, "<=", +1);
        }
#endif
}


