/*
 * Copyright (c) 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * 
 */
/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Debugger
 * FILE:      debuggerStreams.h
 * OVERVIEW:  Header file for Java level debugger code that implements a 
 *            subset of the JPDA.  This file defines the input/output
 *            streams that are used to read/write data to/from the user
 *            level Java debugger
 * AUTHOR:    From JPDA sources/modified by Bill Pittore, Sun Labs
 *=======================================================================*/

/*
 * Data structs
 *
 * NOTE! Individual items in the Packet(s) do not need
 * to be adjusted for endianness. Every field *except* for
 * the PacketData.data field is automatically adjusted
 * by the underlying transport. However, there is no way
 * for the transport to know the format of the raw data,
 * so you must handle this. 
 */

#ifndef __STREAM_H__
#define __STREAM_H__

#define FLAGS_None         ((unsigned char)0x0)
#define FLAGS_Reply        ((unsigned char)0x80)

#define REPLY_NoError      ((short)0x0)

typedef struct PacketStream {
    long numPointers;    /* for pointerlist GC */
    struct Packet *packet;
    struct PacketData *segment;
    long left;
    int error;
} PacketStream, PacketInputStream, PacketOutputStream;

typedef struct PacketData {
    long numPointers;    /* for pointerlist GC */
    unsigned char *data;
    struct PacketData *next;
    long length;
    int index;
} PacketData;

typedef struct CmdPacket {
    long numPointers;    /* for pointerlist GC */
    PacketData *data;
    long id;
    unsigned char flags;
    unsigned char cmdSet;
    unsigned char cmd;
} CmdPacket;

typedef struct ReplyPacket {
    long numPointers;    /* for pointerlist GC */
    PacketData *data;
    long id;
    unsigned char flags;
    short errorCode;
} ReplyPacket;

typedef struct Packet {
    union {
        CmdPacket cmd;
        ReplyPacket reply;
    } type;
} Packet;

typedef union Value {
    bool_t z;
    unsigned char    b;
    char    c;
    short   s;
    long     i;
    union {
        unsigned long    j1;
        unsigned long    j2;
    } u;
    OBJECT  l;
} Value;

#if defined(__i386) || defined(x86) || defined(WINDOWS)

#define HOST_TO_JAVA_CHAR(x) (((x & 0xff) << 8) | ((x >> 8) & (0xff)))
#define HOST_TO_JAVA_SHORT(x) (((x & 0xff) << 8) | ((x >> 8) & (0xff)))
#define HOST_TO_JAVA_INT(x)                                             \
                  ((x << 24) |                                          \
                   ((x & 0x0000ff00) << 8) |                            \
                   ((x & 0x00ff0000) >> 8) |                            \
                   (((unsigned int)(x & 0xff000000)) >> 24))
#define HOST_TO_JAVA_LONG(x)                                            \
                  ((x << 56) |                                          \
                   ((x & 0x000000000000ff00) << 40) |                   \
                   ((x & 0x0000000000ff0000) << 24) |                   \
                   ((x & 0x00000000ff000000) << 8) |                    \
                   ((x & 0x000000ff00000000) >> 8) |                    \
                   ((x & 0x0000ff0000000000) >> 24) |                   \
                   ((x & 0x00ff000000000000) >> 40) |                   \
                   (((ulong64)(x & 0xff00000000000000)) >> 56));

#else 

#define HOST_TO_JAVA_CHAR(x)   (x)
#define HOST_TO_JAVA_SHORT(x)  (x)
#define HOST_TO_JAVA_INT(x)    (x)
#define HOST_TO_JAVA_LONG(x)   (x)
#define HOST_TO_JAVA_FLOAT(x)  (x)
#define HOST_TO_JAVA_DOUBLE(x) (x)

#endif /* i386, i86, Windows */
                             
#define JAVA_TO_HOST_CHAR(x)   HOST_TO_JAVA_CHAR(x)  
#define JAVA_TO_HOST_SHORT(x)  HOST_TO_JAVA_SHORT(x) 
#define JAVA_TO_HOST_INT(x)    HOST_TO_JAVA_INT(x)   
#define JAVA_TO_HOST_LONG(x)   HOST_TO_JAVA_LONG(x)  
#define JAVA_TO_HOST_FLOAT(x)  HOST_TO_JAVA_FLOAT(x) 
#define JAVA_TO_HOST_DOUBLE(x) HOST_TO_JAVA_DOUBLE(x)

/*
 * Input stream defines
 */

void inStream_init(PacketInputStream *stream);

long inStream_id(PacketInputStream *stream);
unsigned char inStream_command(PacketInputStream *stream);

bool_t inStream_readBoolean(PacketInputStream *stream);
unsigned char inStream_readByte(PacketInputStream *stream);
char inStream_readChar(PacketInputStream *stream);
short inStream_readShort(PacketInputStream *stream);
int inStream_readInt(PacketInputStream *stream);
long inStream_readLong(PacketInputStream *stream);
void inStream_readLong64(PacketInputStream *stream, unsigned long *, unsigned long *);
long inStream_readObjectID(PacketInputStream *stream);
FRAME inStream_readFrameID(PacketInputStream *stream);
METHOD inStream_readMethodID(PacketInputStream *stream);
FIELD inStream_readFieldID(PacketInputStream *stream);
long inStream_readLocation(PacketInputStream *stream);

OBJECT inStream_readObjectRef(PacketInputStream *stream);
CLASS inStream_readClassRef(PacketInputStream *stream);
THREAD inStream_readThreadRef(PacketInputStream *stream);
OBJECT inStream_readClassLoaderRef(PacketInputStream *stream);
STRING_INSTANCE inStream_readStringRef(PacketInputStream *stream);
ARRAY inStream_readArrayRef(PacketInputStream *stream);

char *inStream_readString(PacketInputStream *stream);
Value inStream_readValue(PacketInputStream *in, unsigned char *typeKeyPtr);

int inStream_skipBytes(PacketInputStream *stream, int count);

bool_t inStream_endOfInput(PacketInputStream *stream);
int inStream_error(PacketInputStream *stream);

void inStream_destroy(PacketInputStream *stream);

/*
 * Output stream defines
 */

#define INITIAL_SEGMENT_SIZE   256
#define MAX_SEGMENT_SIZE       512
#define SIZEOF_PACKETSTREAM    StructSizeInCells(PacketStream)
#define SIZEOF_PACKETDATA      StructSizeInCells(PacketData)
#define SIZEOF_PACKET          StructSizeInCells(Packet)

void outStream_initCommand(PacketOutputStream *stream, long id, 
    unsigned char flags, unsigned char commandSet, unsigned char command);
void outStream_initReply(PacketOutputStream *stream, long id);

long outStream_id(PacketOutputStream *stream);
unsigned char outStream_command(PacketOutputStream *stream);

int outStream_writeBoolean(PacketOutputStream *stream, bool_t val);
int outStream_writeByte(PacketOutputStream *stream, unsigned char val);
int outStream_writeChar(PacketOutputStream *stream, short val);
int outStream_writeShort(PacketOutputStream *stream, short val);
int outStream_writeInt(PacketOutputStream *stream, int val);
int outStream_writeLong(PacketOutputStream *stream, long val);
int outStream_writeLong64(PacketOutputStream *stream, unsigned long val);
int outStream_writeObjectRef(PacketOutputStream *stream, OBJECT val);
int outStream_writeObjectTag(PacketOutputStream *stream, OBJECT val);
int outStream_writeClassID(PacketOutputStream *stream, CLASS val);
int outStream_writeFrameID(PacketOutputStream *stream, FRAME val);
int outStream_writeMethodID(PacketOutputStream *stream, METHOD val);
int outStream_writeFieldID(PacketOutputStream *stream, FIELD val);
int outStream_writeLocation(PacketOutputStream *stream, BYTE tag, CLASS clazz, METHOD method, unsigned long loc);

#define WRITE_LOCAL_REF(stream, ref) \
    outStream_writeObjectRef(stream, ref)
#define WRITE_GLOBAL_REF(stream, ref) \
    do {                                   \
        outStream_writeObjectRef(stream, ref); \
    } while (0)

int outStream_writeByteArray(PacketOutputStream*stream, int length, unsigned char *bytes);
int outStream_writeString(PacketOutputStream *stream, CONST_CHAR_HANDLE string);
void outStream_writeObjectValue(PacketOutputStream *out, OBJECT object, unsigned char typeTag);
void outStream_writeValue(PacketOutputStream *out, unsigned char typeKey, void *value, void *value2, int writeTag );

int outStream_skipBytes(PacketOutputStream *stream, int count);

int outStream_error(PacketOutputStream *stream);
void outStream_setError(PacketOutputStream *stream, int error);

void outStream_sendReply(PacketOutputStream *stream);
void outStream_sendCommand(PacketOutputStream *stream);

void outStream_destroy(PacketOutputStream *stream);

long uniqueID(void);

#endif /* __STREAM_H__ */

